from PyQt6.QtWidgets import QCheckBox
from trilent.Utility import Reloader, V, Misc
from trilent.Widgets import PositionTypes


class CheckBox(Misc):
    def __init__(self,
                 parent,
                 # Special
                 width: int | str = None,
                 height: int | str = None,
                 command: lambda: ... = None,
                 text: str = None,
                 text_size: int | str = None,
                 font: str = None,  # Add font parameter

                 # Stylesheet
                 checked_color: str = None,
                 unchecked_color: str = None,
                 text_color: str = None,
                 corner_roundness: str | int = None):
        # Reloader setup
        base = \
f"""QCheckBox::indicator:checked {{
    background-color: {V}checked_color{V};
}}
QCheckBox::indicator:unchecked {{
    background-color: {V}unchecked_color{V};
}}
QCheckBox {{
    color: {V}text_color{V};
    font-size: {V}text_size{V};
    font-family: {V}font{V};
}}
QCheckBox::indicator {{
    border-radius: {V}corner_roundness{V};
}}
"""

        self._reloader = Reloader(parent.get_dpi(),
                                  setup_properties={'checked_color': checked_color, 'unchecked_color': unchecked_color,
                                                    'text_color': text_color, 'command': command, 'parent': parent,
                                                    'width': width, 'height': height, 'text': text, 'text_size': text_size, 'font': font, 'corner_roundness': corner_roundness},
                                  process_types={'checked_color': 'color', 'unchecked_color': 'color',
                                                 'text_color': 'color', 'command': 'super raw', 'parent': None,
                                                 'width': 'px-value;int', 'height': 'px-value;int', 'text': None, 'text_size': 'px-value', 'font': 'raw', 'corner_roundness': 'px-value'},
                                  default_values={'checked_color': 'cornflowerblue', 'unchecked_color': 'royalblue',
                                                  'text_color': 'white', 'command': lambda: ..., 'parent': None,
                                                  'width': '1.75 inch', 'height': '0.45 inch', 'text': 'Trilent CheckBox.', 'text_size': '13 px', 'font': 'Arial', 'corner_roundness': '2 px'},
                                  property_types={'checked_color': 'stylesheet', 'unchecked_color': 'stylesheet',
                                                  'text_color': 'stylesheet', 'command': 'access', 'parent': 'access',
                                                  'width': 'special', 'height': 'special', 'text': 'access', 'text_size': 'stylesheet', 'font': 'stylesheet', 'corner_roundness': 'stylesheet'},
                                  special_functions={
                                      'width': lambda v: self._widget.setGeometry(self.x, self.y, v, self.height),
                                      'height': lambda v: self._widget.setGeometry(self.x, self.y, self.width, v)},
                                  base=base)

        # Protected widget specific members
        # noinspection PyProtectedMember
        self._position_self = self._reloader.cp['parent']._position_children
        self._position_children = PositionTypes.NONE
        self._dpi = parent.get_dpi()

        # Qt setup
        self._widget = QCheckBox(self._reloader.cp['text'], self._reloader.cp['parent']._get_holder())  # Use text parameter

        # Setup properties
        self._widget.setStyleSheet(self._reloader.reload())
        self._widget.stateChanged.connect(self._reloader.cp['command'])

        # Check if Box's parent is also a box
        # noinspection PyProtectedMember
        if self._reloader.cp['parent']._position_children == PositionTypes.BOX:
            # noinspection PyProtectedMember
            self._reloader.cp['parent']._widget_box_add(self)

    def is_on(self): return self._widget.isChecked()
    def is_mid(self): return self._widget.isTristate()
    def is_off(self): return not self._widget.isChecked()



from conversa import distance
from functools import lru_cache

trilent_to_conversa = {'inch': 'inches'}


@lru_cache
def get_in_pixels(value: str | int, dpi: int) -> int | float:
    if isinstance(value, int) or isinstance(value, float):
        return value  # Already in px

    value, unit = value.split(' ')[0], value.split(' ')[1] if len(value.split(' ')) == 2 else 'px'

    if unit == 'px':
        return float(value)  # Again, already in px
    else:
        unit_in_inches = distance(float(value), trilent_to_conversa[unit], 'inches')
        return int(unit_in_inches * dpi)



__all__ = ['PositionTypes', 'Box', 'Widget', 'Text', 'Button', 'Entry', 'Slider', 'CheckBox', 'Switch', 'Canvas', 'Image']

from .widget import PositionTypes, Widget
from .box import Box
from .text import Text
from .button import Button
from .entry import Entry
from .slider import Slider
from .check_box import CheckBox
from .switch import Switch
from .canvas import Canvas
from .image import Image

_ = PositionTypes, Box, Widget, Text, Button, Entry, Slider, CheckBox, Switch, Canvas, Image



from typing import Literal
from PyQt6.QtWidgets import QApplication, QMainWindow
from PyQt6.QtCore import QTimer, QTime
from trilent.Utility.unit import get_in_pixels
from trilent.Utility.color import get_as_qt
from trilent.Utility.misc import Misc
from functools import lru_cache
from trilent.Widgets.widget import PositionTypes
from PyQt6.QtCore import Qt


def has_parameter(func, param_names):
    return any(param_name in func.__code__.co_varnames for param_name in param_names)


class Window(Misc):
    _app: QApplication = QApplication([])
    _widget: QMainWindow = QMainWindow()

    def __init__(self,
                 # Utility
                 title: str = 'Trilent Window',
                 width: int | str = '8 inch',
                 height: int | str = '5 inch',
                 x: int | str = '4 inch',
                 y: int | str = '3 inch',
                 # Color
                 background_color='#212121'):

        self._update_timer = None
        self._previous_time = None
        self._position_self = PositionTypes.PLACE
        self._position_children = PositionTypes.PLACE
        self._dpi = self.get_dpi()
        self._update_functions = []
        self.delta = 1

        self._widget.setStyleSheet(f'background-color: {get_as_qt(background_color)};')
        self._widget.setWindowTitle(title)
        self._widget.setGeometry(get_in_pixels(x, self._dpi), get_in_pixels(y, self._dpi), get_in_pixels(width, self._dpi), get_in_pixels(height, self._dpi))

        self._setup_delta_setter()

    # Basic Methods
    def run(self, update=None, start=None, update_speed: int = 0):
        if update is not None:
            if has_parameter(update, ['delta', 'delta_time', 'elapsed_time', 'render_time', 'time_took_to_render']):
                self._update_functions.append(lambda: update(self.delta))
            else:
                self._update_functions.append(update)
        if self._update_functions:
            self._update_timer = QTimer(self._widget)

            def update_func():
                for func in self._update_functions:
                    func()

            self._update_timer.timeout.connect(update_func)
            self._update_timer.start(update_speed)
        if start is not None:
            QTimer(self._widget).singleShot(0, start)

        self._widget.show()
        self._app.exec()

    def attribute(self, name: Literal['opacity', 'transparent', 'show at taskbar'], value, warn=True):
        match name:
            case 'transparent':
                if value:
                    self._widget.setWindowFlags(self._widget.windowFlags() | Qt.WindowType.FramelessWindowHint)
                    self._widget.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground, True)
                else:
                    self._widget.setWindowFlags(self._widget.windowFlags() | Qt.WindowType.Window)
                    self._widget.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground, False)
                    self._widget.show()
            case 'opacity':
                self._log('Opacity may glitch slightly in Windows.', warn)
                self._widget.setWindowOpacity(value / 100)
            case 'show at taskbar':
                if value:
                    self._widget.setWindowFlags(self._widget.windowFlags() | Qt.WindowType.Window)
                else:
                    self._widget.setWindowFlags(self._widget.windowFlags() | Qt.WindowType.Tool)
                self._log('"Show at taskbar" may remove the starting animation of the window', warn)

    @staticmethod
    def _log(string, warn):
        if warn:
            print(string)
            print('To disable this warning, Set warn=False')

    # Utility Methods
    def close(self): self._app.quit()
    @lru_cache
    def get_dpi(self): return int(self._app.primaryScreen().logicalDotsPerInchX())

    def add_update_function(self, func, parse_delta=False):
        if not parse_delta:
            self._update_functions.append(func)
        else:
            self._update_functions.append(lambda: func(self.delta))

    def remove_update_function(self, func):
        self._update_functions = [f for f in self._update_functions if not (callable(f) and f.__closure__ and f.__closure__[0].cell_contents == func)]

    def after(self, ms, func): QTimer(self._widget).singleShot(ms, func)

    def fullscreen(self): self._widget.showFullScreen()
    def normal(self): self._widget.show()
    def minimize(self): self._widget.showMinimized()
    def maximize(self): self._widget.showMaximized()

    def _get_holder(self): return self._widget

    def get_top_parent(self): return self

    def _setup_delta_setter(self):
        self._previous_time = QTime.currentTime()

        def delta_update():
            current_time = QTime.currentTime()
            delta_time = self._previous_time.msecsTo(current_time) / 1000.0
            self._previous_time = current_time
            self.delta = delta_time

        self._update_functions.append(delta_update)


if __name__ == '__main__':
    Window().run()



from functools import lru_cache


def convert_hex_to_rgba(hex_color: str) -> list[int]:
    if hex_color.startswith('rgb(') or hex_color.startswith('rgba('):
        return split_rgb_format(hex_color)

    hex_color += 'FF'

    return list(int((hex_color + 'FF').removeprefix('#')[i:i + 2], 16) for i in (0, 2, 4, 6))


def convert_rgb_to_hex(r, g, b, a=255) -> str:
    return '#{:02x}{:02x}{:02x}{:02x}'.format(r, g, b, a)


def split_rgb_format(rgb_string: str):
    color_channels = [int(channel) for channel in rgb_string.removeprefix('rgb(').removeprefix('rgba(')[:-1].split(', ')]

    if len(color_channels) == 3:
        color_channels.append(255)

    return color_channels


@lru_cache(30)
def gradient_list(color1: str | tuple[int, int, int], color2: str | tuple[int, int, int], steps: int = 10) -> list[str]:
    color1 = convert_hex_to_rgba(color1) if isinstance(color1, str) else color1
    color2 = convert_hex_to_rgba(color2) if isinstance(color2, str) else color2

    steps -= 1

    step_r = (color2[0] - color1[0]) / steps
    step_g = (color2[1] - color1[1]) / steps
    step_b = (color2[2] - color1[2]) / steps
    step_a = (color2[3] - color1[3]) / steps

    gradient = [convert_rgb_to_hex(int(color1[0] + step_r * i), int(color1[1] + step_g * i), int(color1[2] + step_b * i), int(color1[3] + step_a * i)) for i in range(steps + 1)]

    return gradient


def bench():
    gradient = gradient_list('#FF0000', '#0000FF', steps=255)

    print(gradient)


if __name__ == '__main__':
    import cProfile
    cProfile.run('bench()')



import math
from trilent.Utility.gradient_list import gradient_list


def better_range(start_value, end_value):
    if start_value < end_value:
        ranges = list(range(start_value, end_value + 1))
    else:
        ranges = list(range(start_value, end_value, -1))
    return ranges


class Animation:
    def __init__(self, widget, name, end_value, time: float = 0.3, curve='ease in out', animation_finished=lambda: ...):
        # Initials
        self._after_function = None
        self.widget = widget
        self.time = time
        self.name = name
        self.animation_finished = animation_finished

        # Top level window
        self._window = widget.get_top_parent()

        # Defining Values
        if name in ['x','y']:
            self._start_value = getattr(widget, name)
            self._end_value = widget._reloader.alternate_process('px-value;int', end_value)
            self._type = 'px-value;int'
            self._raw_start_value = getattr(widget, name)
            self._raw_end_value = end_value
        else:
            self._start_value = self.widget._reloader.process(self.name, widget._reloader.cp[name])
            self._end_value = widget._reloader.process(name, end_value)
            self._type = widget._reloader.initial_parameters['process_types'][name]
            self._raw_start_value = widget._reloader.cp[name]
            self._raw_end_value = end_value

        self._current_value = self._start_value

        # Animation variables
        self.elapsed_time = 0

        if isinstance(curve, str):
            self._easing_function = getattr(self, curve.replace(' ', '_'))
        elif curve is None:
            self._easing_function = self.ease_in_out_sine
        else:
            self._easing_function = curve

        self.listings = self._generate_listings()

        # Breaker variables
        self._is_stopping = False
        self._stop_elapsed_time = 0
        self._stop_duration = 1

    def start(self):
        if self._start_value != self._end_value:
            self._window.add_update_function(self.update, parse_delta=True)
        else:
            self.animation_finished()

    def update(self, delta):
        if self._is_stopping:
            self._stop_elapsed_time += delta
            proportion = self._easing_function(min(self._stop_elapsed_time / self._stop_duration, 1))
            self.time += proportion * delta

            if self._stop_elapsed_time >= self._stop_duration:
                self._is_stopping = False
                self._window.remove_update_function(self.update)
                self.animation_finished()
                self._after_function()
                return

        self.elapsed_time += delta
        proportion = self._easing_function(min(self.elapsed_time / self.time, 1))

        index = int(proportion * (len(self.listings) - 1))
        index = max(min(index, len(self.listings) - 1), 0)

        self._current_value = self.listings[index]

        if self.name in ['x', 'y']:
            self.widget.set_position(self._current_value, self.widget.y) if self.name == 'x' else self.widget.set_position(self.widget.x, self._current_value)
        else:
            self.widget.set(self.name, self._current_value)

        if self._current_value == self.listings[-1]:
            self._window.remove_update_function(self.update)

            if self.name in ['x', 'y']:
                self.widget.set_position(self._end_value, self.widget.y) if self.name == 'x' else self.widget.set_position(self.widget.x, self._current_value)
            else:
                self.widget.set(self.name, self._end_value if not isinstance(self._end_value, str) else self._end_value.removesuffix('px'))

            self.animation_finished()

    def stop(self, stop_duration=1, after_function=lambda: ...):
        self._is_stopping = True
        self._stop_duration = stop_duration
        self._stop_elapsed_time = 0
        self._after_function = after_function

    def change(self, new_end_value=None, break_speed=None, other_speed=None):
        if new_end_value is None:
            new_end_value = self._current_value
        if break_speed is None:
            break_speed = 0.1
        if other_speed is None:
            other_speed = self.time - self.elapsed_time

        self.stop(break_speed)

        break_speed = max(break_speed, 0.1)

        self._window.after(int((break_speed * 1000)) - 100, lambda: Animation(widget=self.widget, name=self.name, end_value=new_end_value, time=other_speed, curve=self._easing_function, animation_finished=self.animation_finished).start())

    def _generate_listings(self):
        match self._type:
            case 'px-value;int':
                listings = better_range(self._start_value, self._end_value)
            case 'px-value':
                listings = better_range(self.break_px(self._start_value), self.break_px(self._end_value))
            case 'color':
                listings = gradient_list(self._start_value, self._end_value, steps=255)
            case _:
                raise NotImplementedError(f'Animations for "{self.name}" isn\'t supported yet. Type: {self._type}')
        return listings

    @staticmethod
    def break_px(string: str):
        return int(string.removesuffix('px'))

    @staticmethod
    def ease_in(t):
        return t * t * t

    @staticmethod
    def ease_out(t):
        t -= 1
        return t * t * t + 1

    @staticmethod
    def ease_in_out(t):
        t *= 2
        if t < 1:
            return 0.5 * t * t * t
        t -= 2
        return 0.5 * (t * t * t + 2)

    @staticmethod
    def ease_out_back(t, s=1.70158):
        t -= 1
        return t * t * ((s + 1) * t + s) + 1

    @staticmethod
    def ease_in_back(t, s=1.70158):
        return t * t * ((s + 1) * t - s)

    @staticmethod
    def ease_in_out_back(t, s=1.70158):
        t *= 2
        if t < 1:
            return 0.5 * (t * t * ((s + 1) * t - s))
        t -= 2
        return 0.5 * (t * t * ((s + 1) * t + s) + 2)

    @staticmethod
    def ease_in_sine(t):
        return 1 - math.cos((t * math.pi) / 2)

    @staticmethod
    def ease_out_sine(t):
        return math.sin((t * math.pi) / 2)

    @staticmethod
    def ease_in_out_sine(t):
        return -(math.cos(math.pi * t) - 1) / 2

    @staticmethod
    def ease_in_quad(t):
        return t * t

    @staticmethod
    def ease_out_quad(t):
        return t * (2 - t)

    @staticmethod
    def ease_in_out_quad(t):
        return 2 * t * t if t < 0.5 else -1 + (4 - 2 * t) * t

    @staticmethod
    def ease_in_cubic(t):
        return t * t * t

    @staticmethod
    def ease_out_cubic(t):
        return (--t) * t * t + 1

    @staticmethod
    def ease_in_out_cubic(t):
        return 4 * t * t * t if t < 0.5 else (t - 1) * (2 * t - 2) * (2 * t - 2) + 1

    @staticmethod
    def ease_in_quart(t):
        return t * t * t * t

    @staticmethod
    def ease_out_quart(t):
        return 1 - (--t) * t * t * t

    @staticmethod
    def ease_in_out_quart(t):
        return 8 * t * t * t * t if t < 0.5 else 1 - 8 * (--t) * t * t * t

    @staticmethod
    def ease_in_quint(t):
        return t * t * t * t * t

    @staticmethod
    def ease_out_quint(t):
        return 1 + (--t) * t * t * t * t

    @staticmethod
    def ease_in_out_quint(t):
        return 16 * t * t * t * t * t if t < 0.5 else 1 + 16 * (--t) * t * t * t * t

    @staticmethod
    def ease_in_expo(t):
        return 0 if t == 0 else math.pow(2, 10 * (t - 1))

    @staticmethod
    def ease_out_expo(t):
        return 1 if t == 1 else 1 - math.pow(2, -10 * t)

    @staticmethod
    def ease_in_out_expo(t):
        if t == 0:
            return 0
        if t == 1:
            return 1
        return math.pow(2, 10 * (2 * t - 1)) / 2 if t < 0.5 else (2 - math.pow(2, -10 * (2 * t - 1))) / 2

    @staticmethod
    def ease_in_circle(t):
        return 1 - math.sqrt(1 - t * t)

    @staticmethod
    def ease_out_circle(t):
        return math.sqrt(1 - (--t) * t)

    @staticmethod
    def ease_in_out_circle(t):
        return (1 - math.sqrt(1 - 4 * t * t)) / 2 if t < 0.5 else (math.sqrt(1 - (2 * t - 2) * (2 * t - 2)) + 1) / 2



from enum import Enum
from trilent.Animations import Animation

events_dictionary: dict[str, str] = {'wheel': 'wheelEvent',
                                     'drag': 'mouseMoveEvent',
                                     'hover': 'enterEvent',
                                     'unhover': 'leaveEvent',
                                     'clicked': 'mousePressEvent',
                                     'unclicked': 'mouseReleaseEvent'}


class PositionTypes(Enum):
    BOX: str = 'box'
    PLACE: str = 'place'
    NONE: str = 'None specified.'


class Misc:
    def __init__(self):
        self._reloader = None
        self._widget = None

    def place(self, x: int = 100, y: int = 100):
        assert self._reloader.cp['parent']._position_children != PositionTypes.BOX, \
            TypeError("You cant place a widget whose parent is a box. Instead, Its position is automatically handled.")

        geometry = self._widget.geometry()
        self._widget.setGeometry(x, y, geometry.width(), geometry.height())
        self._widget.show()

    def force_place(self, x: int = 100, y: int = 100):
        geometry = self._widget.geometry()
        self._widget.setGeometry(x, y, geometry.width(), geometry.height())
        self._widget.show()

    def set_tooltip(self, tooltip): self._widget.setToolTip(tooltip)
    def get_tooltip(self): self._widget.toolTip()

    def set_position(self, x, y): self._widget.setGeometry(x, y, self.width, self.height)
    def set_size(self, width, height): self._widget.setGeometry(self.x, self.y, width, height)

    def get_position(self):
        geometry = self._widget.geometry()
        return geometry.x(), geometry.y()

    def get_size(self):
        geometry = self._widget.geometry()
        return geometry.width(), geometry.height()

    def change(self, **kwargs):
        for k, v in kwargs.items():
            self._reloader.set(k, v)

        self._widget.setStyleSheet(self._reloader.reload())

    def set(self, name, value):
        self._reloader.set(name, value)

        ss = self._reloader.reload()

        if self._reloader.initial_parameters['property_types'][name] == 'stylesheet':
            self._widget.setStyleSheet(ss)

    def get(self, property_name):
        return self._reloader.cp[property_name]

    @property
    def width(self): return self._widget.width()
    @property
    def height(self): return self._widget.height()
    @property
    def x(self): return self._widget.x()
    @property
    def y(self): return self._widget.y()

    def show(self): self._widget.show()
    def hide(self): self._widget.hide()

    def is_hidden(self): return self._widget.isHidden()
    def is_shown(self): return not self._widget.isHidden()

    def get_top_parent(self): return self.get('parent').get_top_parent()

    def connect(self, name, func): setattr(self._widget, events_dictionary[name], lambda e: func())

    def animate(self, time=0.3, curve=None, animation_finished=lambda: ..., **kwargs):
        for i, (k, v) in enumerate(kwargs.items()):
            Animation(self, k, v, time=time, curve=curve, animation_finished=animation_finished if i == 0 else lambda: ...).start()

    def enable(self): self._widget.setDisabled(False)
    def disable(self): self._widget.setDisabled(True)



from .general import fallback, CACHING_LIMIT, FUNCTIONS, center_y_positions
from functools import lru_cache


@lru_cache(CACHING_LIMIT // FUNCTIONS)
def calculate_end_positions(widths: tuple[int, ...], heights: tuple[int, ...], flex_box_size: tuple[int | str, int | str], wrap: bool, gap: int, vertical_gap: int, side_alignment: str) -> list[list[int]]:
    if side_alignment == 'end':
        x, y = 0, flex_box_size[1]
    else:
        x, y = 0, 0

    positions: list[list[int]] = []
    column_heights: list[int] = []
    column_widths: list[list[int]] = [[]]
    column_widths_map: dict[int, int] = {}
    wraps: int = 1
    total_height: int = 0

    for i, width in enumerate(widths):
        if wrap and x + width > flex_box_size[0] and i != 0:
            x = 0
            if side_alignment == 'end':
                y -= max(column_heights) + vertical_gap
            else:
                y += max(column_heights) + vertical_gap
            total_height += max(column_heights) + vertical_gap

            wraps += 1
            column_heights = []
            column_widths.append([])

        if side_alignment == 'end':
            positions.append([x, y - heights[i]])
        else:
            positions.append([x, y])

        column_heights.append(heights[i])
        column_widths[wraps - 1].append(widths[i] + gap)
        column_widths_map[i] = wraps - 1

        x += widths[i] + gap

    total_height += max(column_heights) + vertical_gap if column_heights else 0

    # Fallback is used to fix a bug, You can ignore this.
    if wraps >= len(widths):
        positions = fallback(positions, heights, vertical_gap, side_alignment, flex_box_size[1])
        positions = [[pos[0], pos[1] - heights[-1]] for pos in positions]

    if side_alignment == 'center':
        positions = center_y_positions(positions, flex_box_size, total_height)

    # End point all x positions and return them
    return [[pos[0] + flex_box_size[0] - sum(column_widths[column_widths_map[i]]) + gap, pos[1]] for i, pos in enumerate(positions)]



from PyQt6.QtWidgets import QSlider
from PyQt6.QtCore import Qt
from trilent.Utility import Reloader, V, Misc
from trilent.Widgets import PositionTypes
from typing import Literal


class Slider(Misc):
    def __init__(self,
                 parent,
                 # Special
                 width: int | str = None,
                 height: int | str = None,
                 command: lambda: ... = None,
                 minimum: int = None,
                 maximum: int = None,

                 # Access
                 quality: str | int = None,

                 # Stylesheet
                 unfilled_color: str = None,
                 button_color: str = None,
                 filled_color: str = None,

                 orientation: Literal['horizontal', 'vertical'] = 'horizontal',
                 corner_roundness: str | int = None,
                 handle_corner_roundness: str | int = None,

                 # Kick starters
                 starting_value: int = 0):

        # Reloader setup
        base = \
f"""QSlider::handle {{
    background-color: {V}button_color{V};
    border-radius: {V}handle_corner_roundness{V};
}}
QSlider::sub-page {{
    background-color: {V}filled_color{V};
    border-radius: {V}corner_roundness{V};
}}
QSlider::add-page {{
    background-color: {V}unfilled_color{V};
    border-radius: {V}corner_roundness{V};
}}
"""

        self._reloader = Reloader(parent.get_dpi(),
                                  setup_properties={'corner_roundness': corner_roundness, 'handle_corner_roundness': handle_corner_roundness, 'minimum': minimum, 'maximum': maximum, 'quality': quality, 'orientation': orientation, 'filled_color': filled_color, 'button_color': button_color, 'command': command, 'unfilled_color': unfilled_color, 'parent': parent, 'width': width, 'height': height,},
                                  process_types={'corner_roundness': 'px-value', 'handle_corner_roundness': 'px-value', 'minimum': 'super raw', 'maximum': 'super raw', 'quality': 'px-value;int', 'orientation': 'raw', 'filled_color': 'color', 'button_color': 'color', 'command': 'super raw', 'unfilled_color': 'color', 'parent': None, 'width': 'px-value;int', 'height': 'px-value;int'},
                                  default_values={'corner_roundness': '0.022 inch', 'handle_corner_roundness': '0.04 inch', 'minimum': 0, 'maximum': 100, 'quality': 1, 'orientation': 'horizontal', 'filled_color': 'cornflowerblue', 'button_color': 'white', 'command': lambda v: ..., 'unfilled_color': 'royalblue', 'parent': None, 'width': '1.75 inch', 'height': '0.45 inch'},
                                  property_types={'corner_roundness': 'stylesheet', 'handle_corner_roundness': 'stylesheet', 'minimum': 'special', 'maximum': 'special', 'quality': 'access', 'orientation': 'special', 'filled_color': 'stylesheet', 'button_color': 'stylesheet', 'command': 'access', 'unfilled_color': 'stylesheet', 'parent': 'access', 'width': 'special', 'height': 'special'},
                                  special_functions={'minimum': lambda v: self._widget.setMinimum(v), 'maximum': lambda v: self._widget.setMaximum(v), 'orientation': lambda v: self._widget.setOrientation(self._orientations[v]), 'width': lambda v: self._widget.setGeometry(self.x, self.y, v, self.height), 'height': lambda v: self._widget.setGeometry(self.x, self.y, self.width, v)},
                                  base=base)

        # Protected widget specific members
        # noinspection PyProtectedMember
        self._position_self = self._reloader.cp['parent']._position_children
        self._position_children = PositionTypes.NONE
        self._dpi = parent.get_dpi()

        # Qt setup
        # noinspection PyProtectedMember
        self._orientations = {'horizontal': Qt.Orientation.Horizontal, 'vertical': Qt.Orientation.Vertical}
        self._widget = QSlider(self._orientations[orientation], self._reloader.cp['parent']._get_holder())

        def slider_command(v: int):
            v = (v // self._reloader.cp['quality']) * self._reloader.cp['quality']
            self._widget.setValue(v)
            self._reloader.cp['command'](v)

        # Setup properties
        self._widget.setValue(starting_value)
        self._widget.setStyleSheet(self._reloader.reload())
        self._widget.valueChanged.connect(slider_command)

        # Check if Box's parent is also a box
        # noinspection PyProtectedMember
        if self._reloader.cp['parent']._position_children == PositionTypes.BOX:
            # noinspection PyProtectedMember
            self._reloader.cp['parent']._widget_box_add(self)



from functools import lru_cache
from typing import Literal

colour_names = {'aliceblue': '#F0F8FF', 'antiquewhite': '#FAEBD7', 'aqua': '#00FFFF', 'aquamarine': '#7FFFD4', 'azure': '#F0FFFF', 'beige': '#F5F5DC', 'bisque': '#FFE4C4', 'black': '#000000', 'blanchedalmond': '#FFEBCD', 'blue': '#0000FF', 'blueviolet': '#8A2BE2', 'brown': '#A52A2A', 'burlywood': '#DEB887', 'cadetblue': '#5F9EA0', 'chartreuse': '#7FFF00', 'chocolate': '#D2691E', 'coral': '#FF7F50', 'cornflowerblue': '#6495ED', 'cornsilk': '#FFF8DC', 'crimson': '#DC143C', 'litecrimson': '#F21642', 'delightcrimson': '#D11339', 'cyan': '#00FFFF', 'darkblue': '#00008B', 'darkcyan': '#008B8B', 'darkgoldenrod': '#B8860B', 'darkgray': '#A9A9A9', 'darkgreen': '#006400', 'darkgrey': '#A9A9A9', 'darkkhaki': '#BDB76B', 'darkmagenta': '#8B008B', 'darkolivegreen': '#556B2F', 'darkorange': '#FF8C00', 'darkorchid': '#9932CC', 'darkred': '#8B0000', 'darksalmon': '#E9967A', 'darkseagreen': '#8FBC8F', 'darkslateblue': '#483D8B', 'darkslategray': '#2F4F4F', 'darkslategrey': '#2F4F4F', 'darkturquoise': '#00CED1', 'darkviolet': '#9400D3', 'deeppink': '#FF1493', 'deepskyblue': '#00BFFF', 'dimgray': '#696969', 'dimgrey': '#696969', 'dodgerblue': '#1E90FF', 'firebrick': '#B22222', 'floralwhite': '#FFFAF0', 'forestgreen': '#228B22', 'fuchsia': '#FF00FF', 'gainsboro': '#DCDCDC', 'ghostwhite': '#F8F8FF', 'gold': '#FFD700', 'goldenrod': '#DAA520', 'gray': '#808080', 'green': '#008000', 'greenyellow': '#ADFF2F', 'grey': '#808080', 'honeydew': '#F0FFF0', 'hotpink': '#FF69B4', 'indianred': '#CD5C5C', 'indigo': '#4B0082', 'ivory': '#FFFFF0', 'khaki': '#F0E68C', 'lavender': '#E6E6FA', 'lavenderblush': '#FFF0F5', 'lawngreen': '#7CFC00', 'lemonchiffon': '#FFFACD', 'lightblue': '#ADD8E6', 'lightcoral': '#F08080', 'lightcyan': '#E0FFFF', 'lightgoldenrodyellow': '#FAFAD2', 'lightgray': '#D3D3D3', 'lightgreen': '#90EE90', 'lightgrey': '#D3D3D3', 'lightpink': '#FFB6C1', 'lightsalmon': '#FFA07A', 'lightseagreen': '#20B2AA', 'lightskyblue': '#87CEFA', 'lightslategray': '#778899', 'lightslategrey': '#778899', 'lightsteelblue': '#B0C4DE', 'lightyellow': '#FFFFE0', 'lime': '#00FF00', 'limegreen': '#32CD32', 'linen': '#FAF0E6', 'magenta': '#FF00FF', 'maroon': '#800000', 'mediumaquamarine': '#66CDAA', 'mediumblue': '#0000CD', 'mediumorchid': '#BA55D3', 'mediumpurple': '#9370DB', 'mediumseagreen': '#3CB371', 'mediumslateblue': '#7B68EE', 'mediumspringgreen': '#00FA9A', 'mediumturquoise': '#48D1CC', 'mediumvioletred': '#C71585', 'midnightblue': '#191970', 'mintcream': '#F5FFFA', 'mistyrose': '#FFE4E1', 'moccasin': '#FFE4B5', 'navajowhite': '#FFDEAD', 'navy': '#000080', 'oldlace': '#FDF5E6', 'olive': '#808000', 'olivedrab': '#6B8E23', 'orange': '#FFA500', 'orangered': '#FF4500', 'orchid': '#DA70D6', 'palegoldenrod': '#EEE8AA', 'palegreen': '#98FB98', 'paleturquoise': '#AFEEEE', 'palevioletred': '#DB7093', 'papayawhip': '#FFEFD5', 'peachpuff': '#FFDAB9', 'peru': '#CD853F', 'pink': '#FFC0CB', 'plum': '#DDA0DD', 'powderblue': '#B0E0E6', 'purple': '#800080', 'red': '#FF0000', 'rosybrown': '#BC8F8F', 'royalblue': '#4169E1', 'saddlebrown': '#8B4513', 'salmon': '#FA8072', 'sandybrown': '#F4A460', 'seagreen': '#2E8B57', 'seashell': '#FFF5EE', 'sienna': '#A0522D', 'silver': '#C0C0C0', 'skyblue': '#87CEEB', 'slateblue': '#6A5ACD', 'slategray': '#708090', 'slategrey': '#708090', 'snow': '#FFFAFA', 'springgreen': '#00FF7F', 'steelblue': '#4682B4', 'tan': '#D2B48C', 'teal': '#008080', 'thistle': '#D8BFD8', 'tomato': '#FF6347', 'turquoise': '#40E0D0', 'violet': '#EE82EE', 'wheat': '#F5DEB3', 'white': '#FFFFFF', 'whitesmoke': '#F5F5F5', 'yellow': '#FFFF00', 'yellowgreen': '#9ACD32', 'xylon': '#0074D9'}
rect_definings = {'horizontal': 'x1: 0, y1: 0, x2: 1, y2: 0',
                  'vertical': 'x1: 0, y1: 0, x2: 0, y2: 1',
                  'diagonal': 'x1: 0, y1: 0, x2: 1, y2: 1',
                  'inverted diagonal': 'x1: 1, y1: 0, x2: 0, y2: 1'}


@lru_cache
def get_hex_colour(string):
    name = string.lower()
    return colour_names[name]


def convert_gradient(value):
    direction: Literal['horizontal', 'vertical', 'diagonal', 'inverted diagonal'] = 'horizontal'
    colors = value.split(' -> ')

    if value.__contains__(' | '):
        direction = value.split(' | ')[-1]
        colors = value.split(' | ')[0].split(' -> ')

    qt_string = f"qlineargradient({rect_definings[direction]}, "

    each_color_increment = 1 / (len(colors) - 1)

    for i, color in enumerate(colors):
        qt_string += f'stop: {each_color_increment * i} {get_as_qt(color)}{', ' if i + 1 != len(colors) else ''}'

    return qt_string + ')'


@lru_cache
def get_color_type(value: str):
    if value == 'transparent':
        color_type = 'special'
    elif value.count('->'):
        color_type = 'gradient'
    elif value.startswith('#') and len(value) == 7:  # 7 because it includes "#"
        color_type = 'hex'
    elif value.startswith('#') and len(value) > 7:
        color_type = 'long-hex'
    elif value.count(',') == 2:
        color_type = 'rgb'
    elif value.count(',') > 2:
        color_type = 'long-rgb'
    else:
        color_type = 'hex'
        value = get_hex_colour(value)
    return value, color_type


def hexa_to_rgba(hexa):
    r = int(hexa[1:3], 16)
    g = int(hexa[3:5], 16)
    b = int(hexa[5:7], 16)
    a = int(hexa[7:9], 16)
    return f'rgba({r}, {g}, {b}, {a})'


def get_as_qt(value):
    value, color_type = get_color_type(value)

    if color_type == 'special':
        return 'transparent'
    elif color_type == 'hex':
        return value
    elif color_type == 'long-hex':
        return hexa_to_rgba(value)
    elif color_type == 'rgb':
        split = value.split(',')
        return f'rgb({split[0]},{split[1]},{split[2]})'
    elif color_type == 'long-rgb':
        split = value.split(',')
        return f'rgba({split[0]},{split[1]},{split[2]},{split[3]})'
    elif color_type == 'gradient':
        return convert_gradient(value)

    raise ValueError(f'Invalid color type: {value}')
